//
// FILE:       query.C
// AUTHORS:    Emanuel G. Noik
//

#include <assert.h>
#include <tk.h>

#include <util/MN_Dictionary.h>
#include <util/MN_Set.h>
#include <util/MN_Point.h>
#include <util/unix.h>
#include <hygraph/Hygraph.h>
#include <server/io.h>
#include <server/server.h>
#include <server/hgInfo.h>

#include "tclTkHInfo.h"
#include "tkHInfo.h"

//------------------------------------------------------------------------------

// Constants.
//
const int nHistBars = 32;

#define NORMAL	0
#define PAINTED	1
#define MASKED	2

//#define Set( &histogram[t][h].masked, h );

//------------------------------------------------------------------------------

// Types.
//

class HistoBar {
public:
   int count;		// number of tuples mapped to this bar
   int nvisible;	// number of tuples shown in current view
   float avg;		// average value over all tuples
   HistoBar( void )
   {
      count = 0;
      nvisible = 0;
      avg = 0;
   };
};

class Histo {
public:
   HistoBar bar[nHistBars];
   int nbars;
   bool inUse;
   float min, max;
   unsigned long int masked;
   unsigned long int painted;
   Histo( void )
   {
      nbars = 0;
      inUse = false;
      min = max = 0;
      masked = 0;
      painted = 0;
   };
};

//------------------------------------------------------------------------------

// Private variables.
//
static char* clrHist  = 0;
static char* clrPaint = 0;
static char* clrMask  = 0;
static Histo histogram[3][nHistBars];
static int lastHisto[3] = { 0, 0, 0 };

//------------------------------------------------------------------------------

inline void Clear( long unsigned int* bm, int i )
{
   (*bm) &= ((long unsigned int)(0xffffffff) - (1 << i));
}

inline void Set( long unsigned int* bm, int i )
{
   (*bm) |= (long unsigned int)(1 << i);
}

inline int IsSet( long unsigned int bm, int i )
{
   return (bm & (long unsigned int)(1 << i)) ? 1 : 0;
}

inline int IsClear( long unsigned int bm, int i )
{
   return (bm & (long unsigned int)(1 << i)) ? 0 : 1;
}

//------------------------------------------------------------------------------

// Set bit 'h' of object 'o' indicating that component 'h' masks object 'o'.
//
inline void SetMask( int t, HGObject* o, int h, int on )
{
   if( on )
      Set( &DQ(mask,o), h );
   else
      Clear( &DQ(mask,o), h );
}

// Set bit 'b' of component 'h' indicating that bucket 'b' is masked.
//
inline void SetMask( int t, int h, int b, int on )
{
   if( on )
      Set( &histogram[t][h].masked, b );
   else
      Clear( &histogram[t][h].masked, b );
}

// Set bit 'b' of component 'h' indicating that bucket 'b' is painted.
//
inline void SetPaint( int t, int h, int b, int on )
{
   if( on )
      Set( &histogram[t][h].painted, b );
   else
      Clear( &histogram[t][h].painted, b );
}

inline int Masked( int t, int h, int b )
{
   return IsSet( histogram[t][h].masked, b );
}

inline int Painted( int t, int h, int b )
{
   return IsSet( histogram[t][h].painted, b );
}

inline int Masked( int t, HGObject* o )
{
   return DQ(mask,o) != 0 ? 1 : 0;
}

//------------------------------------------------------------------------------

int TkHInfo::UpdateHist( int t, int h, int width, int height, int n )
{
   int i, j, rc = ComputeStats( t, h );
   HGObject* o;
   int hist[nHistBars];
   char type = Code( t );

   if( rc != ERR_OK ) return rc;

   histogram[t][h].nbars = n;
   histogram[t][h].inUse = true;
   histogram[t][h].min   =  MAXFLOAT;
   histogram[t][h].max   = -MAXFLOAT;
   lastHisto[t] = Max( lastHisto[t], h+1 );

   for( i = 0; i < n; i++ )
   {
      histogram[t][h].bar[i].count    = 0;
      histogram[t][h].bar[i].nvisible = 0;
      histogram[t][h].bar[i].avg      = 0;
   }

   for( j = 0; (o = Object( t, j )) != NIL; j++ )
   {
      int b = Selects( t, h, o );	// new bucket for object
      int m = Masked( t, h, b );	// is histogram bar masked?
      int p = Painted( t, h, b );	// is histogram bar painted?
      SetMask( t, o, h, m );
      if( DQ(mask,o) != 0 && m != 0 )
         SetState( t, o, MASKED, 1 );
      else if( DQ(mask,o) == 0 && m == 0 )
      {
         SetState( t, o, MASKED, 0 );
         SetState( t, o, PAINTED, p );
      }

      histogram[t][h].bar[b].count++;
      if( ! Masked( t, o ) ) histogram[t][h].bar[b].nvisible++;
      histogram[t][h].bar[b].avg += DQ(val[h],o);
      histogram[t][h].min = Min( histogram[t][h].min, DQ(val[h],o) );
      histogram[t][h].max = Max( histogram[t][h].max, DQ(val[h],o) );
   }

   {
      char xs[64];
      CommaNumber( histogram[t][h].min, xs, 2 );
      sprintf( tmpbuf, "set hmin(.fe.c4.f.%chist%d.c) {%s}", type, h, xs );
      Tcl_GlobalEval( interp, tmpbuf );
      CommaNumber( histogram[t][h].max, xs, 2 );
      sprintf( tmpbuf, "set hmax(.fe.c4.f.%chist%d.c) {%s}", type, h, xs );
      Tcl_GlobalEval( interp, tmpbuf );
   }

   {
      int xs = width  / n;
      //int ys = height / n;
      int max = 0;

      for( i = 0; i < n; i++ ) max = Max( max, histogram[t][h].bar[i].count );
      float s = (float)height / (float)max;
      for( i = 0; i < n; i++ )
      {
         int y = (int)( (float)histogram[t][h].bar[i].count * s );
         hist[i] = histogram[t][h].bar[i].count > 0 ? Max( y, (int)3 ) : 0;
         //y = (int)( (float)histogram[t][h].bar[i].nvisible * s );
         //histB[i] = histogram[t][h].bar[i].nvisible > 0 ? Max( y, (int)3 ) : 0;
         if( histogram[t][h].bar[i].count > 0 )
            histogram[t][h].bar[i].avg /= (float)histogram[t][h].bar[i].count;
      }
      for( i = 0; i < n; i++ )
      {
         sprintf( tmpbuf, ".fe.c4.f.%chist%d.c coords %d %d %d %d %d",
         type, h, i+1, i * xs, height-1, (i+1)*xs, height-hist[i] );
         Tcl_GlobalEval( interp, tmpbuf );
      }
   }

   //EGN? Filter( type );

   sprintf( tmpbuf, "UpdateMappings %c %d", type, h );
   Tcl_GlobalEval( interp, tmpbuf );

   return ERR_OK;
}

//------------------------------------------------------------------------------

// Determine which bucket of histogram 'h' object 'o' belongs to.
//
int TkHInfo::Selects( int t, int h, HGObject* o )
{
   int n = histogram[t][h].nbars;
   return MapNormal( 0, n-1, stat[t][h].Normal( DQ(val[h],o) ) );
}

//------------------------------------------------------------------------------

// Change the visible state of a histogram bar.
//
void TkHInfo::SetState( int t, int h, int b, int state, float x )
{
   if( clrHist == 0 ) LinkVariables();
   char* clr = state == PAINTED ? clrPaint :
               state == MASKED  ? clrMask : clrHist;
   char w[20];
   char type = Code( t );

   sprintf( w, ".fe.c4.f.%chist%d.c", type, h );

   sprintf( tmpbuf, "%s itemconfigure %d -fill %s", w, b+1, clr );
   Tcl_GlobalEval( interp, tmpbuf );

   switch( state )
   {
      case PAINTED:
         {
            char xs[64];
            CommaNumber( x, xs, 2 );
            sprintf( tmpbuf, "set hvalue(%s) {n=%d v=%s}",
               w, histogram[t][h].bar[b].count, xs );
            Tcl_GlobalEval( interp, tmpbuf );
         }
         break;
      case NORMAL:
      case MASKED:
         sprintf( tmpbuf, "set hvalue(%s) {}", w );
         Tcl_GlobalEval( interp, tmpbuf );
         break;
      default:
         break;
   }
}

//------------------------------------------------------------------------------

// Change the state of a node/link.
//
void TkHInfo::SetState( int t, HGObject* o, int state, int on )
{
   if( clrHist == 0 ) LinkVariables();
   HGContainerObj* co = CObject( t, o );
   static int n_masked[2] = { 0, 0 };	// how many objs are masked?

   switch( state )
   {
      case PAINTED:
         if( on && ! DQ(painted,o) )
         {
            if( t == 0 )
               sprintf( tmpbuf,
                  ".canvas itemconfigure %d -outline %s -width 2",
                  o->c->widget, clrPaint );
            else
            {
               sprintf( tmpbuf, ".canvas itemcget %d -fill", co->widget );
               Tcl_GlobalEval( interp, tmpbuf );
               sprintf( DQ(color,o), interp->result );
               sprintf( tmpbuf, ".canvas itemconfigure %d -fill %s",
                        co->widget, clrPaint );
            }
            Tcl_GlobalEval( interp, tmpbuf );
            DQ(painted,o) = true;
         }
         else if( ! on && DQ(painted,o) )
         {
            if( t == 0 )
               sprintf( tmpbuf,
                  ".canvas itemconfigure %d -outline {} -width 1",
                  o->c->widget );
            else
               sprintf( tmpbuf, ".canvas itemconfigure %d -fill %s",
                  co->widget, DQ(color,o) );
            Tcl_GlobalEval( interp, tmpbuf );

            DQ(painted,o) = false;
         }
         break;
      case MASKED:
         if( on && ! DQ(masked,o) )
         {
            if( t == 0 )
            {
               MN_Point d(-10000,-10000,0);
               Translate( o->c, &d );
               //Move( (Vertex*)o, &d );


               DQ(masked,o) = true;
               n_masked[0]++;
               float n = hygraph->V();
               float p = (n - n_masked[0]) / n;
               sprintf( tmpbuf, "SetVisPerc v %g", p );
               Tcl_GlobalEval( interp, tmpbuf );

               o->c->gc.show = false;		// for layout engine
               dirty = true;

               AArc* a;		// also mask incident links
               int j;
               for( j = 0; (a = hygraph->In((Vertex*)o,j)) != NIL; j++ )
                  SetState( 1, a, MASKED, 1 );
               for( j = 0; (a = hygraph->Out((Vertex*)o,j)) != NIL; j++ )
                  SetState( 1, a, MASKED, 1 );

		// EGN: must also mask contained blobs and nodes!!
		// but remember: translate already hides these, so
		// must only set bitmasks?
            }
            else	// t == 1
            {
               sprintf( tmpbuf, ".canvas move %d -10000 -10000", co->widget );
               Tcl_GlobalEval( interp, tmpbuf );
               DQ(masked,o) = true;

               n_masked[1]++;
               float n = hygraph->A();
               float p = (n - n_masked[1]) / n;
               sprintf( tmpbuf, "SetVisPerc a %g", p );
               Tcl_GlobalEval( interp, tmpbuf );

               ((AArc*)o)->ca->gc.show = false;		// for layout engine
               dirty = true;
            }
         }
         else if( ! on && DQ(masked,o) )
         {
            if( t == 0 )
            {
               MN_Point d(10000,10000,0);
               Translate( o->c, &d );
               //Move( (Vertex*)o, &d );

               DQ(masked,o) = false;

               n_masked[0]--;
               float n = hygraph->V();
               float p = (n - n_masked[0]) / n;
               sprintf( tmpbuf, "SetVisPerc v %g", p );
               Tcl_GlobalEval( interp, tmpbuf );

               o->c->gc.show = true;		// for layout engine
               dirty = true;

               AArc* a;			// unmask non-masked incident links
               int j;
               for( j = 0; (a = hygraph->In((Vertex*)o,j)) != NIL; j++ )
                  SetState( 1, a, MASKED, 0 );
               for( j = 0; (a = hygraph->Out((Vertex*)o,j)) != NIL; j++ )
                  SetState( 1, a, MASKED, 0 );
            }
            else	// t == 1
            {
               AArc* a = (AArc*)o;
               if( DQ(mask,o) == 0 && ! DQ(masked,hygraph->From(a))
                                   && ! DQ(masked,hygraph->To(a)) )
               {
		// EGN: should check if graph redrawn; if so, remove
		// mid-points, else keep them ???
                  DumpMP( a->ca, "", a->ca->MP() );

                  DQ(masked,o) = false;

                  n_masked[1]--;
                  float n = hygraph->A();
                  float p = (n - n_masked[1]) / n;
                  sprintf( tmpbuf, "SetVisPerc a %g", p );
                  Tcl_GlobalEval( interp, tmpbuf );

                  ((AArc*)o)->ca->gc.show = true;	// for layout engine
                  dirty = true;
               }
            }
         }
         break;
      default:
         break;
   }
}

//------------------------------------------------------------------------------

// Turn histogram bar and all associated elements ON or OFF.
//
void TkHInfo::Paint( int t, int h, int b, int on )
{
   if( ! Masked( t, h, b ) )
   {
      HGObject* o;
      SetPaint( t, h, b, on );
      SetState( t, h, b, on ? PAINTED : NORMAL, histogram[t][h].bar[b].avg );
      for( int i = 0; (o = Object( t, i )) != NIL; i++ )
         if( Selects( t, h, o )  == b )
            if( ! Masked( t, o ) )
               SetState( t, o, PAINTED, on );
   }
}

//------------------------------------------------------------------------------

// Mask/Show histogram bar and all associated elements.
//
void TkHInfo::Mask( int t, int h, int b, int mode )
{
   int n = histogram[t][h].nbars;
   int on = IsClear( histogram[t][h].masked, b );
					// toggle masked property
   int first = mode & 1 ? 0   : b;	// first affected histogram bar
   int last  = mode & 2 ? n-1 : b;	// last  affected histogram bar

   for( int i = first; i <= last; i++ )
      MaskOne( t, h, i, on, i == b ? PAINTED : NORMAL );
}

//------------------------------------------------------------------------------

// Mask/Show histogram bar and all associated elements.
//
void TkHInfo::MaskOne( int t, int h, int b, int on, int newState )
{
   HGObject* o;
   int p = newState == PAINTED ? 1 : 0;

   SetMask( t, h, b, on );
   SetState( t, h, b, on ? MASKED : newState, histogram[t][h].bar[b].avg );
   for( int i = 0; (o = Object( t, i )) != NIL; i++ )
      if( Selects( t, h, o ) == b )
      {
         SetMask( t, o, h, on );
         if( on )	// masking...
         {
            SetState( t, o, MASKED, on );
         }
         else		// possibly showing
         {
            if( ! Masked( t, o ) )
            {
               SetState( t, o, MASKED, on );
               SetState( t, o, PAINTED, p );
            }
         }
      }
}

//------------------------------------------------------------------------------

// Link/Unlink histogram bars associated with chosen tuple.
//
void TkHInfo::Link( int t, HGObject* o, int on )
{
   SetState( t, o, PAINTED, on );
   for( int h = 0; h < lastHisto[t]; h++ )
      if( histogram[t][h].inUse )
         SetState( t, h, Selects( t, h, o), on ? PAINTED : NORMAL,
                   DQ(val[h],o) );
}

//------------------------------------------------------------------------------

void TkHInfo::LinkVariables( void )
{
   clrHist  = Tcl_GetVar2( interp, "config", "clrHist",   TCL_GLOBAL_ONLY );
   clrPaint = Tcl_GetVar2( interp, "config", "clrSelect", TCL_GLOBAL_ONLY );
   clrMask  = Tcl_GetVar2( interp, "config", "clrHide",   TCL_GLOBAL_ONLY );
}

//------------------------------------------------------------------------------
