//
//  FILE:       tclTkHInfo.C
//  AUTHORS:    Emanuel G. Noik
//

#include <assert.h>
#include <tk.h>

#include <hygraph/Hygraph.h>
#include <util/MN_Dictionary.h>
#include <util/MN_Set.h>
#include <server/io.h>
#include <server/server.h>
#include <server/hgInfo.h>
#include "tclTkHInfo.h"
#include "tcl.h"

//------------------------------------------------------------------------------

#define e(A,i,j) A[i * fd->n + j]

class FloydData : public MN_Element
{
public:
   Hygraph*	h;
   float*	CV;
   float*	CA;
   float*	A;
   int*		PV;
   Arc**	PA;
   int		n;
   //ArcSet	as;
   //VertexSet	vs;

   FloydData( Hygraph* H )
   {
      CV = 0; CA = 0; A = 0; PV = 0; PA = 0;
      Update( H );
   };
   void Update( Hygraph* H )
   {
      Clean();
      if( H )
      {
         n = H->V();
         h  = H;
         CV = new float [n];
         CA = new float [n * n];
         A  = new float [n * n];
         PV = new int   [n * n];
         PA = new Arc*  [n * n];
      }
   };
   void Clean( void ) {
      if( CV ) delete CV;
      if( CA ) delete CA;
      if(  A ) delete  A;
      if( PV ) delete PV;
      if( PA ) delete PA;
   };
   virtual ~FloydData() {
      Clean();
   };
};

//------------------------------------------------------------------------------

static first_time = 1;
static FloydData* fda[32];	// EGN: watch this!

//------------------------------------------------------------------------------

TclTkHInfo::TclTkHInfo( const char* id, Tcl_Interp* intrp ) : HygraphInfo( id )
{
   int i;
   assert( intrp );
   interp = intrp;
   for( i = 0; i < 32; i++ ) matrix[i] = 0;

   if( first_time )
   {
      for( i = 0; i < 32; i++ ) fda[i] = 0;
      first_time = 0;
   }

   htable[0] = new Tcl_HashTable;
   htable[1] = new Tcl_HashTable;
   htable[2] = new Tcl_HashTable;
   Tcl_InitHashTable( htable[0], TCL_ONE_WORD_KEYS );
   Tcl_InitHashTable( htable[1], TCL_ONE_WORD_KEYS );
   Tcl_InitHashTable( htable[2], TCL_ONE_WORD_KEYS );
}

//------------------------------------------------------------------------------

int TclTkHInfo::CreateMatrix( int indx, int size )
{
   if( ! InRange( indx, 0, HG_MaxFIndx ) ) return ERR_BAD_ARG;
   if( matrix[indx] == 0 )
   {
      msize[indx] = size;
      matrix[indx] = new float[size*size];
      mseries[indx] = new MN_Series[size];
   }
   return ERR_OK;
}

int TclTkHInfo::EnlargeMatrix( int indx, int size )
{
   if( ! InRange( indx, 0, HG_MaxFIndx ) ) return ERR_BAD_ARG;
   if( size < msize[indx] ) return ERR_BAD_ARG;
   if( matrix[indx] == 0 ) return ERR_BAD_ARG;

   float* f = new float[size*size];
   MN_Series* s = new MN_Series[size];
   for( int i = 0; i < msize[indx]; i++ )
   {
      for( int j = 0; j < msize[indx]; j++ )
         f[i*size+j] = matrix[indx][i*msize[indx]+j];
      s[i] = mseries[indx][i];
   }

   delete matrix[indx];
   matrix[indx] = f;
   delete mseries[indx];
   mseries[indx] = s;
   msize[indx] = size;

   return ERR_OK;
}

/*
float TclTkHInfo::Dist( int indx, bool normal )
{
   if( ! InRange( indx, 0, HG_MaxFIndx ) ) return ERR_BAD_ARG;
   if( matrix[indx] == 0 ) return 0;
   float dist = matrix[indx][ (fpcursor->i) * msize[indx] + vci ];
   if( normal ) return mseries[indx][fpcursor->i].Normal( dist );
   else return dist;
}
*/

//------------------------------------------------------------------------------

// This is called when we already know that a shortest path exists.
//
static void ConstructPath( FloydData* fd, int i, int j, int indx, int incr )
{
   int k = e(fd->PV,i,j);

   // Direct path: i -> j
   //   mark a[i,j] and v[j]
   //
   if( k == -1 )
   {
      //fd->as.Add( a );
      //fd->vs.Add( (fd->h)->From( a ) );
      //fd->vs.Add( (fd->h)->To( a ) );
      Vertex* v = (fd->h)->V(j);
      Arc* a = e(fd->PA,i,j);
      DQ(npaths[indx],a) += incr;
      DQ(npaths[indx],v) += incr;
   }
   // Intermediate path: i -> k -> j
   //
   else
   {
      ConstructPath( fd, i, k, indx, incr );
      ConstructPath( fd, k, j, indx, incr );
   }
}

//------------------------------------------------------------------------------

//  Given n x 1 array CV of vertex costs and n x n array CA of arc costs,
//  produce n x n array A of path lengths, n x n array PV of intermediate
//  vertices along shortest paths, and n x n array PA of intermediate arcs.
//
static void floyd( FloydData* fd )
{
   int i, j, k;

   for( i = 0; i < fd->n; i++ )
      for( j = 0; j < fd->n; j++ )
      {
         e(fd->A,i,j) = e(fd->CA,i,j)           + fd->CV[i] + fd->CV[j];
         e(fd->PV,i,j) = -1;
      }

   for( i = 0; i < fd->n; i++ )
      e(fd->A,i,i) = /*0*/  fd->CV[i];

   for( k = 0; k < fd->n; k++ )
   {
      for( i = 0; i < fd->n; i++ )
      {
         for( j = 0; j < fd->n; j++ )
         {
            if( e(fd->A,i,k) + e(fd->A,k,j)    - fd->CV[k]    < e(fd->A,i,j) )
            {
               e(fd->A,i,j) = e(fd->A,i,k) + e(fd->A,k,j)    - fd->CV[k];
               e(fd->PV,i,j) = k;
               e(fd->PA,i,j) = e(fd->PA,k,j);
            }
         }
      }
   }
}

//------------------------------------------------------------------------------

static void InitArc( Tcl_Interp* interp, FloydData* fd, Vertex* f, Vertex* t,
Arc* a, Direction d, char* expr )
{
   int I = DQ(i,f);	//hygraph->V( f );
   int J = DQ(i,t);	//hygraph->V( t );

   if( I < 0 || J < 0 ) return;

   float weight;
   {
      double d;
      //stack.Set( a );
      Tcl_ExprDouble( interp, expr, &d );
      weight = Abs( d );
   }

   if( weight < e(fd->CA,I,J) )
   {
      if( d == Direction_Fwd || d == Direction_None )
      {
         e(fd->CA,I,J) = weight;
         e(fd->PA,I,J) = a;
      }
      if( d == Direction_Rev || d == Direction_None )
      {
         e(fd->CA,J,I) = weight;
         e(fd->PA,J,I) = a;
      }
   }
}

int TclTkHInfo::InitMinWtPaths( int indx, Direction d, int type,
char* vvar, char* avar, char* vexpr, char* aexpr )
{
   int i, j;

   if( ! InRange( indx, 0, HG_MaxFIndx ) ) return ERR_BAD_ARG;
   if( ! InRange( type, 1, 2 ) ) return ERR_BAD_ARG;

   if( fda[indx] == 0 )
      fda[indx] = new FloydData( hygraph );
   else
      fda[indx]->Update( hygraph );

   if( fda[indx]->n == 0 ) return -1;

   FloydData* fd = fda[indx];

   {
      //Tcl_SetVar( interp, vvar, "0", 0 );
      Vertex* v;
      //stack.Push( 0 );
      for( i = 0; (v = hygraph->V( i )) != NIL; i++ )
      {
         //stack.Set( v );
         char buf[8];
         sprintf( buf, "%d", (int)v );
         Tcl_SetVar( interp, vvar, buf, 0 );
         double d;
         Tcl_ExprDouble( interp, vexpr, &d );
         fd->CV[i] = (float)Abs( d );
         DQ(i,v) = i;
      }
      //stack.Pop();
      Tcl_UnsetVar( interp, vvar, 0 );
   }

   for( i = 0; i < hygraph->V(); i++ )
      for( j = 0; j < hygraph->V(); j++ )
         e(fd->CA,i,j) = MAXFLOAT;

   //Tcl_SetVar( interp, avar, "0", 0 );
   //stack.Push( type );

   if( type == 1 )
   {
      AArc* a;
      for( i = 0; (a = hygraph->A( i )) != NIL; i++ )
      {
         Vertex* f = hygraph->From( a );
         Vertex* t = hygraph->To( a );
         char buf[8];
         sprintf( buf, "%d", (int)a );
         Tcl_SetVar( interp, avar, buf, 0 );
         InitArc( interp, fd, f, t, a, d, aexpr );
      }
   }
   else		// type == 2
   {
      IArc* a;
      for( i = 0; (a = hygraph->I( i )) != NIL; i++ )
      {
         Vertex* f = hygraph->From( a );
         Vertex* t = hygraph->To( a );
         char buf[8];
         sprintf( buf, "%d", (int)a );
         Tcl_SetVar( interp, avar, buf, 0 );
         InitArc( interp, fd, f, t, a, d, aexpr );
      }
   }

   //stack.Pop();
   Tcl_UnsetVar( interp, avar, 0 );

   floyd( fd );

   return indx;
}

//------------------------------------------------------------------------------

float TclTkHInfo::MinWtPaths( int indx, Vertex* u, Vertex* v )
{
   if( ! InRange( indx, 0, HG_MaxFIndx ) ) return ERR_BAD_ARG;
   FloydData* fd = fda[indx];
   if( fd == NIL ) return 0;

   int ii = DQ(i,u);		//hygraph->V( u );
   int ij = DQ(i,v);		//hygraph->V( v );

   if( ii < 0 || ij < 0 )
      return -1;
   else
   {
      float x = e(fd->A,ii,ij);
      return (x > MAXFLOAT/2) ? -1 : x;
   }
}

//------------------------------------------------------------------------------

int TclTkHInfo::MinWtPaths( int indx )
{
   if( ! InRange( indx, 0, HG_MaxFIndx ) ) return ERR_BAD_ARG;
   FloydData* fd = fda[indx];
   if( fd == NIL ) return ERR_NO_ELT;

   Vertex *u, *v;

   //fd->as.Clear();
   //fd->vs.Clear();

   // EGN:  this should happen exactly once for every
   // path source insertion/deletion rather than all at the same time
   // every time...
   int i;
   {  Vertex* v;
      for( i = 0; (v = (fd->h)->V(i)) != NIL; i++ ) DQ(npaths[indx],v) = 0;
   }
   {  AArc* a;
      for( i = 0; (a = (fd->h)->A(i)) != NIL; i++ ) DQ(npaths[indx],a) = 0;
   }
   {  IArc* a;
      for( i = 0; (a = (fd->h)->I(i)) != NIL; i++ ) DQ(npaths[indx],a) = 0;
   }

   for( i = 0; (u = sources[i]) != NIL; i++ )
   {
      int iu = DQ(i,u);		//hygraph->V( u );
      if( iu < 0 ) return ERR_NO_ELT;
      for( int j = 0; (v = sinks[j]) != NIL; j++ )
      {
         int iv = DQ(i,v);		//hygraph->V( v );
         if( iv < 0 ) return ERR_NO_ELT;
         if( e(fd->A,iu,iv) < MAXFLOAT )
         {
            ConstructPath( fd, iu, iv, indx, 1 ); // Construct path finds all
            DQ(npaths[indx],(fd->h)->V(iu))++;    // but the initial source
         }
      }
   }

   return ERR_OK;
}

//------------------------------------------------------------------------------

int TclTkHInfo::UpdateEndPt( Vertex* u, bool source, int incr )
{
   if( incr == 1 )
   {
      if( source )
         sources.Add( u );
      else
         sinks.Add( u );
   }
   else
   {
      if( source )
         sources.Remove( u );
      else
         sinks.Remove( u );
   }

   for( int i = 0; i < HG_MaxFIndx; i++ )
   {
      if( fda[i] != NIL )
      {
         int rc;
         if( source ) rc = UpdateSource( i, u, incr );
         else rc = UpdateSink( i, u, incr );
         if( rc != ERR_OK ) return rc;
      }
   }
   return ERR_OK;
}

//------------------------------------------------------------------------------

int TclTkHInfo::UpdateSource( int indx, Vertex* u, int incr )
{
   if( ! InRange( indx, 0, HG_MaxFIndx ) ) return ERR_BAD_ARG;
   FloydData* fd = fda[indx];
   if( fd == NIL ) return ERR_NO_ELT;

   Vertex* v;

   int iu = DQ(i,u);
   if( iu < 0 ) return ERR_NO_ELT;
   for( int i = 0; (v = sinks[i]) != NIL; i++ )
   {
      int iv = DQ(i,v);
      if( iv < 0 ) return ERR_NO_ELT;
      if( e(fd->A,iu,iv) < MAXFLOAT )
      {
         ConstructPath( fd, iu, iv, indx, incr ); // Construct path finds all
         DQ(npaths[indx],(fd->h)->V(iu)) += incr; // but the initial source
      }
   }
   return ERR_OK;
}

//------------------------------------------------------------------------------

int TclTkHInfo::UpdateSink( int indx, Vertex* v, int incr )
{
   if( ! InRange( indx, 0, HG_MaxFIndx ) ) return ERR_BAD_ARG;
   FloydData* fd = fda[indx];
   if( fd == NIL ) return ERR_NO_ELT;

   Vertex* u;

   int iv = DQ(i,v);
   if( iv < 0 ) return ERR_NO_ELT;
   for( int i = 0; (u = sources[i]) != NIL; i++ )
   {
      int iu = DQ(i,u);
      if( iu < 0 ) return ERR_NO_ELT;
      if( e(fd->A,iu,iv) < MAXFLOAT )
      {
         ConstructPath( fd, iu, iv, indx, incr ); // Construct path finds all
         DQ(npaths[indx],(fd->h)->V(iu)) += incr; // but the initial source
      }
   }
   return ERR_OK;
}

//------------------------------------------------------------------------------

int TclTkHInfo::OnPath( HGObject* o, int indx, int /*type*/ )
{
   if( ! InRange( indx, 0, HG_MaxFIndx ) ) return ERR_BAD_ARG;
   FloydData* fd = fda[indx];
   if( fd == NIL ) return 0;
   //if(      type == 0 ) return fd->vs.Member( o ) ? 1 : 0;
   //else if( type == 1 ) return fd->as.Member( o ) ? 1 : 0;
   //else return 0;
   return DQ(npaths[indx],o);
}

//------------------------------------------------------------------------------

/*
float MinWtPaths( int indx, VertexSet* sources, Vertex* v )
{
   if( ! InRange( indx, 0, HG_MaxFIndx ) ) return ERR_BAD_ARG;
   FloydData* fd = fda[indx];
   if( fd == NIL ) return 0;

   float wt = MAXFLOAT;
   int ij = fd->V->Index( v );
   if( ij < 0 ) return MAXFLOAT;

   Vertex *u;
   for( int i = 0; (u = (*sources)[i]) != NIL; i++ )
   {
      int ii = fd->V->Index( u );
      if( ii < 0 ) return MAXFLOAT;
      wt = Min( wt, e(fd->A,ii,ij) );
   }

   return wt;
}
*/

//------------------------------------------------------------------------------

int TclTkHInfo::RegExpr( int type, int id, char* expr, int indx )
{
   if( ! InRange( indx, 0, HG_MaxFIndx ) ) return ERR_BAD_ARG;

   //if( re_comp( expr ) == 0 )
   Tcl_RegExp regexp = Tcl_RegExpCompile( interp, expr );
   if( regexp != NULL )
   {
      Vertex* v;
      AArc* a;
      Blob* b;
      int i;

      switch( type )
      {
         case 0:
            for( i = 0; (v = hygraph->V( i )) != NIL; i++ )
               DQ(val[indx],v) =
               //re_exec( id ? hygraph->Id(v) : hygraph->Label(v) ) ? 1 : 0;
               Tcl_RegExpExec( interp, regexp, id ?
                  hygraph->Id(v) : hygraph->Label(v), 0 ) ? 1 : 0;
            break;
         case 1:
            for( i = 0; (a = hygraph->A(i)) != NIL; i++ )
               DQ(val[indx],a) =
               //re_exec( id ? hygraph->Id(a) : hygraph->Label(a) ) ? 1 : 0;
               Tcl_RegExpExec( interp, regexp, id ?
                  hygraph->Id(a) : hygraph->Label(a), 0 ) ? 1 : 0;
            break;
         case 2:
            for( i = 0; (b = hygraph->B(i)) != NIL; i++ )
               DQ(val[indx],v) =
               //re_exec( id ? hygraph->Id(b) : hygraph->Label(b) ) ? 1 : 0;
               Tcl_RegExpExec( interp, regexp, id ?
                  hygraph->Id(b) : hygraph->Label(b), 0 ) ? 1 : 0;
            break;
         default:
            return ERR_BAD_ARG;
            break;
      }
   }
   return ERR_OK;
}

//------------------------------------------------------------------------------

int TclTkHInfo::ComputeStats( int type, int i )
{
   HGObject* o;
   int j;
   
   stat[type][i].Clear();

   for( j = 0; (o = Object( type, j )) != NIL; j++ )
      stat[type][i].Add( DQ(val[i],o) );

   sdirty[type][i] = 0;

   return ERR_OK;
}
 
//------------------------------------------------------------------------------
