/* 
 * tkCanvSsbitmap.c --
 *
 *	This file implements scalable sbitmap items for canvas widgets.
 */

#include <stdio.h>
#include "tkInt.h"
#include "tkPort.h"
#include "tkCanvas.h"

/*
 * The structure below defines the record for each sbitmap item.
 */

typedef struct SBitmapItem  {
    Tk_Item header;		/* Generic stuff that's the same for all
				 * types.  MUST BE FIRST IN STRUCTURE. */
    double bbox[4];		/* Coordinates for sbitmap */
    Pixmap bitmap;		/* SBitmap to display in window. */
    XColor *fgColor;		/* Foreground color to use for sbitmap. */
    XColor *bgColor;		/* Background color to use for sbitmap. */
    XColor *fillColor;		/* Fill color for rectangular region. */
    XColor *outlineColor;	/* Border color to use for sbitmap. */
    int width;			/* Width of border */
    GC gc;			/* Graphics context to use for drawing
				 * sbitmap on screen. */
    GC fillGC;
    GC outlineGC;
} SBitmapItem;

/*
 * Information used for parsing configuration specs:
 */

static Tk_ConfigSpec configSpecs[] = {
    {TK_CONFIG_COLOR, "-background", (char *) NULL, (char *) NULL,
	(char *) NULL, Tk_Offset(SBitmapItem, bgColor), TK_CONFIG_NULL_OK},
    {TK_CONFIG_BITMAP, "-bitmap", (char *) NULL, (char *) NULL,
	(char *) NULL, Tk_Offset(SBitmapItem, bitmap), TK_CONFIG_NULL_OK},
    {TK_CONFIG_COLOR, "-fill", (char *) NULL, (char *) NULL,
	(char *) NULL, Tk_Offset(SBitmapItem, fillColor), TK_CONFIG_NULL_OK},
    {TK_CONFIG_COLOR, "-foreground", (char *) NULL, (char *) NULL,
	"black", Tk_Offset(SBitmapItem, fgColor), 0},
    {TK_CONFIG_COLOR, "-outline", (char *) NULL, (char *) NULL,
	(char *) NULL, Tk_Offset(SBitmapItem, outlineColor), TK_CONFIG_NULL_OK},
    {TK_CONFIG_CUSTOM, "-tags", (char *) NULL, (char *) NULL,
	(char *) NULL, 0, TK_CONFIG_NULL_OK, &tk_CanvasTagsOption},
    {TK_CONFIG_PIXELS, "-width", (char *) NULL, (char *) NULL,
        "1", Tk_Offset(SBitmapItem, width), TK_CONFIG_DONT_SET_DEFAULT},
    {TK_CONFIG_END, (char *) NULL, (char *) NULL, (char *) NULL,
	(char *) NULL, 0, 0}
};

/*
 * Prototypes for procedures defined in this file:
 */

static int		SBitmapCoords _ANSI_ARGS_((Tcl_Interp *interp,
			    Tk_Canvas canvas, Tk_Item *itemPtr, int argc,
			    char **argv));
static int		SBitmapToArea _ANSI_ARGS_((Tk_Canvas canvas,
			    Tk_Item *itemPtr, double *rectPtr));
static double		SBitmapToPoint _ANSI_ARGS_((Tk_Canvas canvas,
			    Tk_Item *itemPtr, double *coordPtr));
static int		SBitmapToPostscript _ANSI_ARGS_((Tcl_Interp *interp,
			    Tk_Canvas canvas, Tk_Item *itemPtr, int prepass));
static void		ComputeSBitmapBbox _ANSI_ARGS_((Tk_Canvas canvas,
			    SBitmapItem *bmapPtr));
static int		ConfigureSBitmap _ANSI_ARGS_((Tcl_Interp *interp,
			    Tk_Canvas canvas, Tk_Item *itemPtr, int argc,
			    char **argv, int flags));
static int		CreateSBitmap _ANSI_ARGS_((Tcl_Interp *interp,
			    Tk_Canvas canvas, struct Tk_Item *itemPtr,
			    int argc, char **argv));
static void		DeleteSBitmap _ANSI_ARGS_((Tk_Canvas canvas,
			    Tk_Item *itemPtr, Display *display));
static void		DisplaySBitmap _ANSI_ARGS_((Tk_Canvas canvas,
			    Tk_Item *itemPtr, Display *display, Drawable dst,
			    int x, int y, int width, int height));
static void		ScaleSBitmap _ANSI_ARGS_((Tk_Canvas canvas,
			    Tk_Item *itemPtr, double originX, double originY,
			    double scaleX, double scaleY));
static void		TranslateSBitmap _ANSI_ARGS_((Tk_Canvas canvas,
			    Tk_Item *itemPtr, double deltaX, double deltaY));

/*
 * The structures below defines the sbitmap item type in terms of
 * procedures that can be invoked by generic item code.
 */

Tk_ItemType tkSBitmapType = {
    "sbitmap",				/* name */
    sizeof(SBitmapItem),		/* itemSize */
    CreateSBitmap,			/* createProc */
    configSpecs,			/* configSpecs */
    ConfigureSBitmap,			/* configureProc */
    SBitmapCoords,			/* coordProc */
    DeleteSBitmap,			/* deleteProc */
    DisplaySBitmap,			/* displayProc */
    0,					/* alwaysRedraw */
    SBitmapToPoint,			/* pointProc */
    SBitmapToArea,			/* areaProc */
    SBitmapToPostscript,		/* postscriptProc */
    ScaleSBitmap,			/* scaleProc */
    TranslateSBitmap,			/* translateProc */
    (Tk_ItemIndexProc *) NULL,		/* indexProc */
    (Tk_ItemCursorProc *) NULL,		/* icursorProc */
    (Tk_ItemSelectionProc *) NULL,	/* selectionProc */
    (Tk_ItemInsertProc *) NULL,		/* insertProc */
    (Tk_ItemDCharsProc *) NULL,		/* dTextProc */
    (Tk_ItemType *) NULL		/* nextPtr */
};

/*
 *--------------------------------------------------------------
 *
 * CreateSBitmap --
 *
 *	This procedure is invoked to create a new sbitmap
 *	item in a canvas.
 *
 * Results:
 *	A standard Tcl return value.  If an error occurred in
 *	creating the item, then an error message is left in
 *	interp->result;  in this case itemPtr is left uninitialized,
 *	so it can be safely freed by the caller.
 *
 * Side effects:
 *	A new sbitmap item is created.
 *
 *--------------------------------------------------------------
 */

static int
CreateSBitmap(
    Tcl_Interp *interp,			/* Interpreter for error reporting. */
    Tk_Canvas canvas,			/* Canvas to hold new item. */
    Tk_Item *itemPtr,			/* Record to hold new item;  header */
					/* has been initialized by caller. */
    int argc,				/* Number of arguments in argv. */
    char *argv[] )			/* Arguments describing rectangle. */
{
    SBitmapItem *bmapPtr = (SBitmapItem *) itemPtr;

    if (argc < 4) {
	Tcl_AppendResult(interp, "wrong # args:  should be \"",
		Tk_PathName(Tk_CanvasTkwin(canvas)), "\" create ",
		itemPtr->typePtr->name, " x1 y1 x2 y2 ?options?",
		(char *) NULL);
	return TCL_ERROR;
    }

    /*
     * Initialize item's record.
     */

    bmapPtr->bitmap = None;
    bmapPtr->fillColor = NULL;
    bmapPtr->outlineColor = NULL;
    bmapPtr->fgColor = NULL;
    bmapPtr->bgColor = NULL;
    bmapPtr->gc = None;
    bmapPtr->fillGC = None;
    bmapPtr->outlineGC = None;
    bmapPtr->width = 1;

    /*
     * Process the arguments to fill in the item record.
     */

    if (
       (Tk_CanvasGetCoord(interp, canvas, argv[0], &bmapPtr->bbox[0]) != TCL_OK)
    || (Tk_CanvasGetCoord(interp, canvas, argv[1], &bmapPtr->bbox[1]) != TCL_OK)
    || (Tk_CanvasGetCoord(interp, canvas, argv[2], &bmapPtr->bbox[2]) != TCL_OK)
    || (Tk_CanvasGetCoord(interp, canvas, argv[3], &bmapPtr->bbox[3]) != TCL_OK)    ) {
	return TCL_ERROR;
    }

    if (ConfigureSBitmap(interp, canvas, itemPtr, argc-4, argv+4, 0) != TCL_OK) {
	DeleteSBitmap(canvas, itemPtr, Tk_Display(Tk_CanvasTkwin(canvas)));
	return TCL_ERROR;
    }
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * SBitmapCoords --
 *
 *	This procedure is invoked to process the "coords" widget
 *	command on sbitmap items.  See the user documentation for
 *	details on what it does.
 *
 * Results:
 *	Returns TCL_OK or TCL_ERROR, and sets interp->result.
 *
 * Side effects:
 *	The coordinates for the given item may be changed.
 *
 *--------------------------------------------------------------
 */

static int
SBitmapCoords(
    Tcl_Interp *interp,			/* Used for error reporting. */
    Tk_Canvas canvas,			/* Canvas containing item. */
    Tk_Item *itemPtr,			/* Item whose coordinates are to be
					 * read or modified. */
    int argc,				/* Number of coordinates supplied in
					 * argv. */
    char *argv[] )			/* Array of coordinates: x1, y1,
					 * x2, y2, ... */
{
    SBitmapItem *bmapPtr = (SBitmapItem *) itemPtr;
    char c[4][TCL_DOUBLE_SPACE];

    if (argc == 0) {
	Tcl_PrintDouble(interp, bmapPtr->bbox[0], c[0]);
	Tcl_PrintDouble(interp, bmapPtr->bbox[1], c[1]);
	Tcl_PrintDouble(interp, bmapPtr->bbox[2], c[2]);
	Tcl_PrintDouble(interp, bmapPtr->bbox[3], c[3]);
	Tcl_AppendResult(interp, c[0], " ", c[1], " ", c[2], " ", c[3],
           (char *) NULL);
    } else if (argc == 4) {
	if ((Tk_CanvasGetCoord(interp, canvas, argv[0], &bmapPtr->bbox[0])
             != TCL_OK)
         || (Tk_CanvasGetCoord(interp, canvas, argv[1], &bmapPtr->bbox[1])
             != TCL_OK)
         || (Tk_CanvasGetCoord(interp, canvas, argv[2], &bmapPtr->bbox[2])
             != TCL_OK)
         || (Tk_CanvasGetCoord(interp, canvas, argv[3], &bmapPtr->bbox[3])
             != TCL_OK)

        ) {
	    return TCL_ERROR;
	}
	ComputeSBitmapBbox(canvas, bmapPtr);
    } else {
	sprintf(interp->result,
		"wrong # coordinates:  expected 0 or 4, got %d", argc);
	return TCL_ERROR;
    }
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * ConfigureSBitmap --
 *
 *	This procedure is invoked to configure various aspects
 *	of a sbitmap item, such as its anchor position.
 *
 * Results:
 *	A standard Tcl result code.  If an error occurs, then
 *	an error message is left in interp->result.
 *
 * Side effects:
 *	Configuration information may be set for itemPtr.
 *
 *--------------------------------------------------------------
 */

static int
ConfigureSBitmap(
    Tcl_Interp *interp,		/* Used for error reporting. */
    Tk_Canvas canvas,		/* Canvas containing itemPtr. */
    Tk_Item *itemPtr,		/* SBitmap item to reconfigure. */
    int argc,			/* Number of elements in argv.  */
    char *argv[],		/* Arguments describing things to configure. */
    int flags )			/* Flags to pass to Tk_ConfigureWidget. */
{
    SBitmapItem *bmapPtr = (SBitmapItem *) itemPtr;
    XGCValues gcValues;
    GC newGC;
    Tk_Window tkwin;
    unsigned long mask;

    tkwin = Tk_CanvasTkwin(canvas);
    if (Tk_ConfigureWidget(interp, tkwin, configSpecs, argc, argv,
	    (char *) bmapPtr, flags) != TCL_OK) {
	return TCL_ERROR;
    }

    /*
     * A few of the options require additional processing, such as those
     * that determine the graphics context.
     */

    if (bmapPtr->width < 1) {
        bmapPtr->width = 1;
    }

    if( bmapPtr->outlineColor == NULL ) {
       newGC = None;
    } else {
       gcValues.foreground = bmapPtr->outlineColor->pixel;
       gcValues.cap_style = CapProjecting;
       gcValues.line_width = bmapPtr->width;
       mask = GCForeground|GCCapStyle|GCLineWidth;
       newGC = Tk_GetGC(tkwin, mask, &gcValues);
    }
    if (bmapPtr->outlineGC != None) {
       Tk_FreeGC(Tk_Display(tkwin), bmapPtr->outlineGC);
    }
    bmapPtr->outlineGC = newGC;

    if (bmapPtr->fillColor == NULL) {
        newGC = None;
    } else {
        gcValues.foreground = bmapPtr->fillColor->pixel;
        mask = GCForeground;
        newGC = Tk_GetGC(tkwin, mask, &gcValues);
    }
    if (bmapPtr->fillGC != None) {
        Tk_FreeGC(Tk_Display(tkwin), bmapPtr->fillGC);
    }
    bmapPtr->fillGC = newGC;

    gcValues.foreground = bmapPtr->fgColor->pixel;
    mask = GCForeground;
    if (bmapPtr->bgColor != NULL) {
	gcValues.background = bmapPtr->bgColor->pixel;
	mask |= GCBackground;
    } else {
	gcValues.clip_mask = bmapPtr->bitmap;
	mask |= GCClipMask;
    }
    newGC = Tk_GetGC(tkwin, mask, &gcValues);
    if (bmapPtr->gc != None) {
	Tk_FreeGC(Tk_Display(tkwin), bmapPtr->gc);
    }
    bmapPtr->gc = newGC;

    ComputeSBitmapBbox(canvas, bmapPtr);

    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * DeleteSBitmap --
 *
 *	This procedure is called to clean up the data structure
 *	associated with a sbitmap item.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Resources associated with itemPtr are released.
 *
 *--------------------------------------------------------------
 */

static void
DeleteSBitmap(
    Tk_Canvas canvas,			/* Info about overall canvas widget. */
    Tk_Item *itemPtr,			/* Item that is being deleted. */
    Display *display )			/* Display containing window for
					 * canvas. */
{
    SBitmapItem *bmapPtr = (SBitmapItem *) itemPtr;

    if (bmapPtr->bitmap != None) {
	Tk_FreeBitmap(display, bmapPtr->bitmap);
    }
    if (bmapPtr->fgColor != NULL) {
	Tk_FreeColor(bmapPtr->fgColor);
    }
    if (bmapPtr->bgColor != NULL) {
	Tk_FreeColor(bmapPtr->bgColor);
    }
    if (bmapPtr->outlineColor != NULL) {
	Tk_FreeColor(bmapPtr->outlineColor);
    }
    if (bmapPtr->fillColor != NULL) {
	Tk_FreeColor(bmapPtr->fillColor);
    }
    if (bmapPtr->gc != NULL) {
	Tk_FreeGC(display, bmapPtr->gc);
    }
}

/*
 *--------------------------------------------------------------
 *
 * ComputeSBitmapBbox --
 *
 *	This procedure is invoked to compute the bounding box of
 *	all the pixels that may be drawn as part of a sbitmap item.
 *	This procedure is where the child sbitmap's placement is
 *	computed.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The fields bbox[i] are updated in the header for itemPtr.
 *
 *--------------------------------------------------------------
 */

static void
ComputeSBitmapBbox(
    Tk_Canvas canvas,			/* Canvas that contains item. */
    SBitmapItem *bmapPtr )		/* Item whose bbox is to be
					 * recomputed. */
{
    int bloat;

    if (bmapPtr->bbox[0] > bmapPtr->bbox[2]) {
       double tmp;
       tmp = bmapPtr->bbox[2];
       bmapPtr->bbox[2] = bmapPtr->bbox[0];
       bmapPtr->bbox[0] = tmp;
    }
    if (bmapPtr->bbox[1] > bmapPtr->bbox[3]) {
       double tmp;
       tmp = bmapPtr->bbox[3];
       bmapPtr->bbox[3] = bmapPtr->bbox[1];
       bmapPtr->bbox[1] = tmp;
    }

    if (bmapPtr->outlineColor == NULL) {
        bloat = 0;
    } else {
        bloat = (bmapPtr->width+1)/2;
    }
    bmapPtr->header.x1 = (int)(bmapPtr->bbox[0] - bloat + 0.5);
    bmapPtr->header.y1 = (int)(bmapPtr->bbox[1] - bloat + 0.5);
    bmapPtr->header.x2 = (int)(bmapPtr->bbox[2] + bloat + 0.5);
    bmapPtr->header.y2 = (int)(bmapPtr->bbox[3] + bloat + 0.5);
}

/*
 *--------------------------------------------------------------
 *
 * DisplaySBitmap --
 *
 *	This procedure is invoked to draw a sbitmap item in a given
 *	drawable.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	ItemPtr is drawn in drawable using the transformation
 *	information in canvas.
 *
 *--------------------------------------------------------------
 */

static void
DisplaySBitmap(
    Tk_Canvas canvas,			/* Canvas that contains item. */
    Tk_Item *itemPtr,			/* Item to be displayed. */
    Display *display,			/* Display on which to draw item. */
    Drawable drawable,			/* Pixmap or window in which to draw
					 * item. */
    int x, int y,
    int width, int height )		/* Describes region of canvas that
					 * must be redisplayed (not used). */
{
    SBitmapItem *bmapPtr = (SBitmapItem *) itemPtr;
    int bmapX, bmapY, bmapWidth, bmapHeight;
    short drawableX, drawableY;

    short x1, y1, x2, y2;

    /*
     * Compute the screen coordinates of the bounding box for the item.
     * Make sure that the bbox is at least one pixel large, since some
     * X servers will die if it isn't.
     */

    Tk_CanvasDrawableCoords(canvas, bmapPtr->bbox[0], bmapPtr->bbox[1],
            &x1, &y1);
    Tk_CanvasDrawableCoords(canvas, bmapPtr->bbox[2], bmapPtr->bbox[3],
            &x2, &y2);
    if (x2 <= x1) {
        x2 = x1+1;
    }
    if (y2 <= y1) {
        y2 = y1+1;
    }

    /*
     * Display filled rectangle first, then draw sbitmap on top.
     */
     if (bmapPtr->fillGC != None) {
        XFillRectangle(display, drawable, bmapPtr->fillGC,
           x1, y1, (unsigned) (x2-x1), (unsigned) (y2-y1));
     }
     if (bmapPtr->outlineGC != None) {
        XDrawRectangle(display, drawable, bmapPtr->outlineGC,
           x1, y1, (unsigned) (x2-x1), (unsigned) (y2-y1));
     }

    /*
     * If the area being displayed doesn't cover the whole sbitmap,
     * then only redisplay the part of the sbitmap that needs
     * redisplay.
     */

    if (bmapPtr->bitmap != None) {
	if (x > bmapPtr->header.x1) {
	    bmapX = x - (int)(bmapPtr->bbox[0]+0.5);
	    bmapWidth = (int)(bmapPtr->bbox[2]+0.5) - x;
	} else {
	    bmapX = 0;
	    if ((x+width) < bmapPtr->bbox[2]) {
		bmapWidth = x + width - (int)(bmapPtr->bbox[0]+0.5);
	    } else {
		bmapWidth = (int)(bmapPtr->bbox[2]+0.5) -
                            (int)(bmapPtr->bbox[0]+0.5);
	    }
	}
	if (y > bmapPtr->header.y1) {
	    bmapY = y - (int)(bmapPtr->bbox[1]+0.5);
	    bmapHeight = (int)(bmapPtr->bbox[3]+0.5) - y;
	} else {
	    bmapY = 0;
	    if ((y+height) < bmapPtr->bbox[3]) {
		bmapHeight = y + height - (int)(bmapPtr->bbox[1]+0.5);
	    } else {
		bmapHeight = (int)(bmapPtr->bbox[3]+0.5) -
                             (int)(bmapPtr->bbox[1]+0.5);
	    }
	}
        int bloat = (bmapPtr->width+1)/2;

	Tk_CanvasDrawableCoords(canvas,
		(double) (bmapPtr->bbox[0] + bloat + bmapX),
		(double) (bmapPtr->bbox[1] + bloat + bmapY),
		&drawableX, &drawableY);

        bmapWidth  -= bloat * 2;
        bmapHeight -= bloat * 2;

	/*
	 * Must modify the mask origin within the graphics context
	 * to line up with the sbitmap's origin (in order to make
	 * sbitmaps with "-background {}" work right).
	 */
 
	XSetClipOrigin(display, bmapPtr->gc, drawableX - bmapX,
		drawableY - bmapY);
	XCopyPlane(display, bmapPtr->bitmap, drawable,
		bmapPtr->gc, bmapX, bmapY, (unsigned int) bmapWidth,
		(unsigned int) bmapHeight, drawableX, drawableY, 1);
    }
}

/*
 *--------------------------------------------------------------
 *
 * SBitmapToPoint --
 *
 *	Computes the distance from a given point to a given
 *	rectangle, in canvas units.
 *
 * Results:
 *	The return value is 0 if the point whose x and y coordinates
 *	are coordPtr[0] and coordPtr[1] is inside the sbitmap.  If the
 *	point isn't inside the sbitmap then the return value is the
 *	distance from the point to the sbitmap.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

static double
SBitmapToPoint(
    Tk_Canvas canvas,		/* Canvas containing item. */
    Tk_Item *itemPtr,		/* Item to check against point. */
    double *coordPtr )		/* Pointer to x and y coordinates. */
{
    SBitmapItem *bmapPtr = (SBitmapItem *) itemPtr;
    double x1, x2, y1, y2, xDiff, yDiff, inc;

    x1 = bmapPtr->bbox[0];
    y1 = bmapPtr->bbox[1];
    x2 = bmapPtr->bbox[2];
    y2 = bmapPtr->bbox[3];

    if (bmapPtr->outlineGC != None) {
        inc = bmapPtr->width/2.0;
        x1 -= inc;
        y1 -= inc;
        x2 += inc;
        y2 += inc;
    }

    if (coordPtr[0] < x1) {
	xDiff = x1 - coordPtr[0];
    } else if (coordPtr[0] > x2)  {
	xDiff = coordPtr[0] - x2;
    } else {
	xDiff = 0;
    }

    if (coordPtr[1] < y1) {
	yDiff = y1 - coordPtr[1];
    } else if (coordPtr[1] > y2)  {
	yDiff = coordPtr[1] - y2;
    } else {
	yDiff = 0;
    }

    return hypot(xDiff, yDiff);
}

/*
 *--------------------------------------------------------------
 *
 * SBitmapToArea --
 *
 *	This procedure is called to determine whether an item
 *	lies entirely inside, entirely outside, or overlapping
 *	a given rectangle.
 *
 * Results:
 *	-1 is returned if the item is entirely outside the area
 *	given by rectPtr, 0 if it overlaps, and 1 if it is entirely
 *	inside the given area.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

static int
SBitmapToArea(
    Tk_Canvas canvas,		/* Canvas containing item. */
    Tk_Item *itemPtr,		/* Item to check against rectangle. */
    double *rectPtr )		/* Pointer to array of four coordinates
				 * (x1, y1, x2, y2) describing rectangular
				 * area.  */
{
    SBitmapItem *bmapPtr = (SBitmapItem *) itemPtr;
    double halfWidth = bmapPtr->width/2.0;
    if (bmapPtr->outlineGC == None) {
        halfWidth = 0.0;
    }

    if ((rectPtr[2] <= (bmapPtr->header.x1 - halfWidth))
	    || (rectPtr[0] >= (bmapPtr->header.x2 + halfWidth))
	    || (rectPtr[3] <= (bmapPtr->header.y1 - halfWidth))
	    || (rectPtr[1] >= (bmapPtr->header.y2 + halfWidth))) {
	return -1;
    }
    if ((rectPtr[0] <= (bmapPtr->header.x1 - halfWidth))
	    && (rectPtr[1] <= (bmapPtr->header.y1 - halfWidth))
	    && (rectPtr[2] >= (bmapPtr->header.x2 + halfWidth))
	    && (rectPtr[3] >= (bmapPtr->header.y2 + halfWidth))) {
	return 1;
    }
    return 0;
}

/*
 *--------------------------------------------------------------
 *
 * ScaleSBitmap --
 *
 *	This procedure is invoked to rescale a sbitmap item in a
 *	canvas.  It is one of the standard item procedures for
 *	sbitmap items, and is invoked by the generic canvas code.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The item referred to by itemPtr is rescaled so that the
 *	following transformation is applied to all point coordinates:
 *		x' = originX + scaleX*(x-originX)
 *		y' = originY + scaleY*(y-originY)
 *
 *--------------------------------------------------------------
 */

static void
ScaleSBitmap(
    Tk_Canvas canvas,			/* Canvas containing rectangle. */
    Tk_Item *itemPtr,			/* Rectangle to be scaled. */
    double originX, double originY,	/* Origin about which to scale item. */
    double scaleX,			/* Amount to scale in X direction. */
    double scaleY )			/* Amount to scale in Y direction. */
{
    SBitmapItem *bmapPtr = (SBitmapItem *) itemPtr;

    bmapPtr->bbox[0] = originX + scaleX*(bmapPtr->bbox[0] - originX);
    bmapPtr->bbox[1] = originY + scaleY*(bmapPtr->bbox[1] - originY);
    bmapPtr->bbox[2] = originX + scaleX*(bmapPtr->bbox[2] - originX);
    bmapPtr->bbox[3] = originY + scaleY*(bmapPtr->bbox[3] - originY);
    ComputeSBitmapBbox(canvas, bmapPtr);
}

/*
 *--------------------------------------------------------------
 *
 * TranslateSBitmap --
 *
 *	This procedure is called to move an item by a given amount.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The position of the item is offset by (xDelta, yDelta), and
 *	the bounding box is updated in the generic part of the item
 *	structure.
 *
 *--------------------------------------------------------------
 */

static void
TranslateSBitmap(
    Tk_Canvas canvas,			/* Canvas containing item. */
    Tk_Item *itemPtr,			/* Item that is being moved. */
    double deltaX, double deltaY )	/* Amount by which item is to be
					 * moved. */
{
    SBitmapItem *bmapPtr = (SBitmapItem *) itemPtr;

    bmapPtr->bbox[0] += deltaX;
    bmapPtr->bbox[1] += deltaY;
    bmapPtr->bbox[2] += deltaX;
    bmapPtr->bbox[3] += deltaY;
    ComputeSBitmapBbox(canvas, bmapPtr);
}

/*
 *--------------------------------------------------------------
 *
 * SBitmapToPostscript --
 *
 *	This procedure is called to generate Postscript for
 *	sbitmap items.
 *
 * Results:
 *	The return value is a standard Tcl result.  If an error
 *	occurs in generating Postscript then an error message is
 *	left in interp->result, replacing whatever used to be there.
 *	If no error occurs, then Postscript for the item is appended
 *	to the result.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

static int
SBitmapToPostscript(
    Tcl_Interp *interp,			/* Leave Postscript or error message
					 * here. */
    Tk_Canvas canvas,			/* Information about overall canvas. */
    Tk_Item *itemPtr,			/* Item for which Postscript is
					 * wanted. */
    int prepass )			/* 1 means this is a prepass to
					 * collect font information;  0 means
					 * final Postscript is being created. */
{
    char pathCmd[500], string[100];
    SBitmapItem *bmapPtr = (SBitmapItem *) itemPtr;
    double x, y, y1, y2;
    int width, height, rowsAtOnce, rowsThisTime;
    int curRow;
    char buffer[200];

   /*
    * First dump the rectangular region.
    */
    y1 = Tk_CanvasPsY(canvas, bmapPtr->bbox[1]);
    y2 = Tk_CanvasPsY(canvas, bmapPtr->bbox[3]);

    sprintf(pathCmd, "%.15g %.15g moveto %.15g 0 rlineto 0 %.15g rlineto %.15g 0 rlineto closepath\n",
       bmapPtr->bbox[0], y1,
       bmapPtr->bbox[2]-bmapPtr->bbox[0], y2-y1,
       bmapPtr->bbox[0]-bmapPtr->bbox[2]);

    if (bmapPtr->fillColor != NULL) {
       Tcl_AppendResult(interp, pathCmd, (char *) NULL);
       if (Tk_CanvasPsColor(interp, canvas, bmapPtr->fillColor) != TCL_OK) {
          return TCL_ERROR;
       }
       Tcl_AppendResult(interp, "fill\n", (char *) NULL);
    }

    if (bmapPtr->outlineColor != NULL) {
       Tcl_AppendResult(interp, pathCmd, (char *) NULL);
       sprintf(string, "%d setlinewidth", bmapPtr->width);
       Tcl_AppendResult(interp, string,
          " 0 setlinejoin 2 setlinecap\n", (char *) NULL);
       if (Tk_CanvasPsColor(interp, canvas, bmapPtr->outlineColor) != TCL_OK) {
           return TCL_ERROR;
       }
       Tcl_AppendResult(interp, "stroke\n", (char *) NULL);
    }

    if (bmapPtr->bitmap == None) {
       return TCL_OK;
    }

    /*
     * Compute the coordinates of the top-left corner of the sbitmap.
     */

    x = bmapPtr->bbox[0];
    y = Tk_CanvasPsY(canvas, bmapPtr->bbox[1]);
    Tk_SizeOfBitmap(Tk_Display(Tk_CanvasTkwin(canvas)), bmapPtr->bitmap,
	    &width, &height);

    if (width > bmapPtr->bbox[2] - bmapPtr->bbox[0])
       width = (int)(bmapPtr->bbox[2] - bmapPtr->bbox[0]);
    if (height > bmapPtr->bbox[3] - bmapPtr->bbox[1])
       height = (int)(bmapPtr->bbox[3] - bmapPtr->bbox[1]);

    /*
     * Color the background, if there is one.
     */

    if (bmapPtr->bgColor != NULL) {
	sprintf(buffer,
		"%.15g %.15g moveto %d 0 rlineto 0 %d rlineto %d %s\n",
		x+1, y-height-1, width, height, -width, "0 rlineto closepath");
	Tcl_AppendResult(interp, buffer, (char *) NULL);
	if (Tk_CanvasPsColor(interp, canvas, bmapPtr->bgColor) != TCL_OK) {
	    return TCL_ERROR;
	}
	Tcl_AppendResult(interp, "fill\n", (char *) NULL);
    }

    /*
     * Draw the sbitmap, if there is a foreground color.  If the sbitmap
     * is very large, then chop it up into multiple sbitmaps, each
     * consisting of one or more rows.  This is needed because Postscript
     * can't handle single strings longer than 64 KBytes long.
     */

    if (bmapPtr->fgColor != NULL) {
	if (Tk_CanvasPsColor(interp, canvas, bmapPtr->fgColor) != TCL_OK) {
	    return TCL_ERROR;
	}
	if (width > 60000) {
	    Tcl_ResetResult(interp);
	    Tcl_AppendResult(interp, "can't generate Postscript",
		    " for sbitmaps more than 60000 pixels wide",
		    (char *) NULL);
	    return TCL_ERROR;
	}
	rowsAtOnce = 60000/width;
	if (rowsAtOnce < 1) {
	    rowsAtOnce = 1;
	}
	sprintf(buffer, "%.15g %.15g translate\n", x+1, y-1 /*+height*/ );
	Tcl_AppendResult(interp, buffer, (char *) NULL);
	for (curRow = 0; curRow < height; curRow += rowsAtOnce) {
	    rowsThisTime = rowsAtOnce;
	    if (rowsThisTime > (height - curRow)) {
		rowsThisTime = height - curRow;
	    }
	    sprintf(buffer, "0 -%.15g translate\n%d %d true matrix {\n",
		    (double) rowsThisTime, width, rowsThisTime);
	    Tcl_AppendResult(interp, buffer, (char *) NULL);
	    if (Tk_CanvasPsBitmap(interp, canvas, bmapPtr->bitmap,
		    0, curRow, width, rowsThisTime) != TCL_OK) {
		return TCL_ERROR;
	    }
	    Tcl_AppendResult(interp, "\n} imagemask\n", (char *) NULL);
        }
    }
    return TCL_OK;
}
